from __future__ import annotations

from pylav.exceptions.base import PyLavException


class PyLavInvalidArgumentsException(PyLavException):
    """Base Exception for when invalid arguments are passed to a method"""


class PyLavNotInitializedException(PyLavException):
    """Raised when the library is not initialized"""


class AnotherClientAlreadyRegisteredException(PyLavException):
    """Another client has already been registered"""


class CogAlreadyRegisteredException(PyLavException):
    """Raised when a cog is already registered"""


class CogHasBeenRegisteredException(PyLavException):
    """Raised when a cog is registered"""



from __future__ import annotations

from dataclasses import dataclass

from pylav.compat import json
from pylav.constants.builtin_nodes import BUNDLED_NODES_IDS_HOST_MAPPING
from pylav.constants.config import JAVA_EXECUTABLE
from pylav.constants.node import NODE_DEFAULT_SETTINGS
from pylav.constants.node_features import SUPPORTED_FEATURES, SUPPORTED_SOURCES
from pylav.extension.bundled_node.utils import get_jar_ram_actual
from pylav.helpers.singleton import SingletonCachedByKey
from pylav.storage.database.cache.decodators import maybe_cached
from pylav.storage.database.cache.model import CachedModel
from pylav.storage.database.tables.nodes import NodeRow
from pylav.type_hints.dict_typing import JSON_DICT_TYPE


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True, frozen=True)
class Node(CachedModel, metaclass=SingletonCachedByKey):
    id: int

    def get_cache_key(self) -> str:
        return f"{self.id}"

    @maybe_cached
    async def exists(self) -> bool:
        """Check if the node exists in the database.

        Returns
        -------
        bool
            Whether the node exists in the database.
        """
        return await NodeRow.exists().where(NodeRow.id == self.id)

    async def delete(self) -> None:
        """Delete the node from the database"""
        await NodeRow.delete().where(NodeRow.id == self.id)
        await self.invalidate_cache()

    @maybe_cached
    async def fetch_all(self) -> JSON_DICT_TYPE:
        data = await NodeRow.select().where(NodeRow.id == self.id).first().output(load_json=True, nested=True)
        return data or {
            "id": self.id,
            "name": NodeRow.name.default,
            "ssl": NodeRow.ssl.default,
            "resume_timeout": NodeRow.resume_timeout.default,
            "reconnect_attempts": NodeRow.reconnect_attempts.default,
            "search_only": NodeRow.search_only.default,
            "managed": NodeRow.managed.default,
            "extras": json.loads(NodeRow.extras.default),
            "yaml": json.loads(NodeRow.yaml.default),
            "disabled_sources": NodeRow.disabled_sources.default,
        }

    @maybe_cached
    async def fetch_name(self) -> str | None:
        """Fetch the node from the database.

        Returns
        -------
        str
            The node's name.
        """
        data = (
            await NodeRow.select(NodeRow.name).where(NodeRow.id == self.id).first().output(load_json=True, nested=True)
        )
        return data["name"] if data else None

    async def update_name(self, name: str) -> None:
        """Update the node's name in the database"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await NodeRow.raw(
            """
            INSERT INTO node (id, name) VALUES ({}, {})
            ON CONFLICT (id) DO UPDATE SET name = excluded.name;
            """,
            self.id,
            name,
        )
        await self.update_cache((self.fetch_name, name), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_ssl(self) -> bool:
        """Fetch the node's ssl setting from the database.

        Returns
        -------
        bool
            The node's ssl setting.
        """
        data = (
            await NodeRow.select(NodeRow.ssl).where(NodeRow.id == self.id).first().output(load_json=True, nested=True)
        )
        return data["ssl"] if data else NodeRow.ssl.default

    async def update_ssl(self, ssl: bool) -> None:
        """Update the node's ssl setting in the database"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await NodeRow.raw(
            """
            INSERT INTO node (id, ssl) VALUES ({}, {})
            ON CONFLICT (id) DO UPDATE SET ssl = excluded.ssl;
            """,
            self.id,
            ssl,
        )
        await self.update_cache((self.fetch_ssl, ssl), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_resume_timeout(self) -> int:
        """Fetch the node's resume timeout from the database.

        Returns
        -------
        int
            The node's resume timeout.
        """
        data = (
            await NodeRow.select(NodeRow.resume_timeout)
            .where(NodeRow.id == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        return data["resume_timeout"] if data else NodeRow.resume_timeout.default

    async def update_resume_timeout(self, resume_timeout: int) -> None:
        """Update the node's resume timeout in the database"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await NodeRow.raw(
            """
            INSERT INTO node (id, resume_timeout) VALUES ({}, {})
            ON CONFLICT (id) DO UPDATE SET resume_timeout = excluded.resume_timeout;
            """,
            self.id,
            resume_timeout,
        )
        await self.update_cache((self.fetch_resume_timeout, resume_timeout), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_reconnect_attempts(self) -> int:
        """Fetch the node's reconnect attempts from the database.

        Returns
        -------
        int
            The node's reconnect attempts.
        """
        data = (
            await NodeRow.select(NodeRow.reconnect_attempts)
            .where(NodeRow.id == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        return data["reconnect_attempts"] if data else NodeRow.reconnect_attempts.default

    async def update_reconnect_attempts(self, reconnect_attempts: int) -> None:
        """Update the node's reconnect attempts in the database"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await NodeRow.raw(
            """
            INSERT INTO node (id, reconnect_attempts) VALUES ({}, {})
            ON CONFLICT (id) DO UPDATE SET reconnect_attempts = excluded.reconnect_attempts;
            """,
            self.id,
            reconnect_attempts,
        )
        await self.update_cache((self.fetch_reconnect_attempts, reconnect_attempts), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_search_only(self) -> bool:
        """Fetch the node's search only setting from the database.

        Returns
        -------
        bool
            The node's search only setting.
        """
        data = (
            await NodeRow.select(NodeRow.search_only)
            .where(NodeRow.id == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        return data["search_only"] if data else NodeRow.search_only.default

    async def update_search_only(self, search_only: bool) -> None:
        """Update the node's search only setting in the database"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await NodeRow.raw(
            """
            INSERT INTO node (id, search_only) VALUES ({}, {})
            ON CONFLICT (id) DO UPDATE SET search_only = excluded.search_only;
            """,
            self.id,
            search_only,
        )
        await self.update_cache((self.fetch_search_only, search_only), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_managed(self) -> bool:
        """Fetch the node's managed setting from the database.

        Returns
        -------
        bool
            The node's managed setting.
        """
        data = (
            await NodeRow.select(NodeRow.managed)
            .where(NodeRow.id == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        return data["managed"] if data else NodeRow.managed.default

    async def update_managed(self, managed: bool) -> None:
        """Update the node's managed setting in the database"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await NodeRow.raw(
            """
            INSERT INTO node (id, managed) VALUES ({}, {})
            ON CONFLICT (id) DO UPDATE SET managed = excluded.managed;
            """,
            self.id,
            managed,
        )
        await self.update_cache((self.fetch_managed, managed), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_extras(self) -> JSON_DICT_TYPE:
        """Fetch the node's extras from the database.

        Returns
        -------
        dict
            The node's extras.
        """
        data = (
            await NodeRow.select(NodeRow.extras)
            .where(NodeRow.id == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        return data["extras"] if data else json.loads(NodeRow.extras.default)

    async def update_extras(self, extras: JSON_DICT_TYPE) -> None:
        """Update the node's extras in the database"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await NodeRow.raw(
            """
            INSERT INTO node (id, extras) VALUES ({}, {})
            ON CONFLICT (id) DO UPDATE SET extras = excluded.extras;
            """,
            self.id,
            json.dumps(extras),
        )
        await self.update_cache((self.fetch_extras, extras), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_yaml(self) -> JSON_DICT_TYPE:
        """Fetch the node's yaml from the database.

        Returns
        -------
        dict
            The node's yaml.
        """
        data = (
            await NodeRow.select(NodeRow.yaml).where(NodeRow.id == self.id).first().output(load_json=True, nested=True)
        )
        return data["yaml"] if data else json.loads(NodeRow.yaml.default)

    async def update_yaml(self, yaml_data: JSON_DICT_TYPE) -> None:
        """Update the node's yaml in the database"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await NodeRow.raw(
            """
            INSERT INTO node (id, yaml) VALUES ({}, {})
            ON CONFLICT (id) DO UPDATE SET yaml = excluded.yaml;
            """,
            self.id,
            json.dumps(yaml_data),
        )
        await self.update_cache((self.fetch_yaml, yaml_data), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_disabled_sources(self) -> list[str]:
        """Fetch the node's disabled sources from the database.

        Returns
        -------
        list[str]
            The node's disabled sources.
        """
        data = (
            await NodeRow.select(NodeRow.disabled_sources)
            .where(NodeRow.id == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        return data["disabled_sources"] if data else NodeRow.disabled_sources.default

    async def update_disabled_sources(self, disabled_sources: list[str]) -> None:
        """Update the node's disabled sources in the database"""
        source = set(map(str.strip, map(str.lower, disabled_sources)))
        intersection = list(source & SUPPORTED_SOURCES.union(SUPPORTED_FEATURES))
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await NodeRow.raw(
            """
            INSERT INTO node (id, disabled_sources) VALUES ({}, {})
            ON CONFLICT (id) DO UPDATE SET disabled_sources = excluded.disabled_sources;
            """,
            self.id,
            intersection,
        )
        await self.update_cache((self.fetch_disabled_sources, intersection), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    async def add_to_disabled_sources(self, source: str) -> None:
        """Add a source to the node's disabled sources in the database"""
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await NodeRow.raw(
            """
            INSERT INTO node (id, disabled_sources) VALUES ({}, {})
            ON CONFLICT (id)
            DO UPDATE SET disabled_sources = ARRAY_CAT(node.disabled_sources, EXCLUDED.disabled_sources);
            """,
            self.id,
            [source],
        )
        await self.update_cache((self.exists, True))
        await self.invalidate_cache(self.fetch_all, self.fetch_disabled_sources)

    async def remove_from_disabled_sources(self, source: str) -> None:
        """Remove a source from the node's disabled sources in the database"""
        # TODO: When piccolo add support to more Array operations replace with ORM
        await NodeRow.raw(
            """UPDATE node SET disabled_sources = array_remove(disabled_sources, {}) WHERE id = {}""",
            source,
            self.id,
        )
        await self.update_cache((self.exists, True))
        await self.invalidate_cache(self.fetch_all, self.fetch_disabled_sources)

    async def bulk_add_to_disabled_sources(self, sources: list[str]) -> None:
        """Add sources to the node's disabled sources in the database"""
        source = set(map(str.strip, map(str.lower, [sources])))
        intersection = list(source & SUPPORTED_SOURCES.union(SUPPORTED_FEATURES))
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await NodeRow.raw(
            """
            INSERT INTO node (id, disabled_sources) VALUES ({}, {})
            ON CONFLICT (id) DO UPDATE SET disabled_sources = disabled_sources || EXCLUDED.disabled_sources;
            """,
            self.id,
            intersection,
        )
        await self.update_cache((self.exists, True))
        await self.invalidate_cache(self.fetch_all, self.fetch_disabled_sources)

    async def bulk_remove_from_disabled_sources(self, sources: list[str]) -> None:
        """Remove sources from the node's disabled sources in the database"""
        if not sources:
            return
        for source in sources:
            await self.remove_from_disabled_sources(source)

    async def bulk_update(
        self,
        host: str,
        port: int,
        password: str,
        resume_timeout: int = 60,
        name: str | None = None,
        reconnect_attempts: int = -1,
        ssl: bool = False,
        search_only: bool = False,
        managed: bool = False,
        extras: JSON_DICT_TYPE | None = None,
        yaml: JSON_DICT_TYPE | None = None,
        disabled_sources: list[str] | None = None,
    ) -> None:
        """Update the node's data in the database"""
        yaml_data = yaml or {"server": {}, "lavalink": {"server": {}}}
        yaml_data["server"]["address"] = host  # type: ignore
        yaml_data["server"]["port"] = port  # type: ignore
        yaml_data["lavalink"]["server"]["password"] = password
        if disabled_sources is None:
            disabled_sources = []
        if extras is None:
            extras = {}
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await NodeRow.raw(
            """
            INSERT INTO node
            (id,
            name,
            ssl,
            resume_key,
            resume_timeout,
            reconnect_attempts,
            search_only,
            managed,
            disabled_sources,
            extras,
            yaml)
            VALUES ({}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {})
            ON CONFLICT (id)
            DO UPDATE
              SET name = excluded.name,
              ssl = excluded.ssl,
              resume_key = excluded.resume_key,
              resume_timeout = excluded.resume_timeout,
              reconnect_attempts = excluded.reconnect_attempts,
              search_only = excluded.search_only,
              managed = excluded.managed,
              disabled_sources = excluded.disabled_sources,
              extras = excluded.extras,
              yaml = excluded.yaml;
            """,
            self.id,
            name,
            ssl,
            None,
            resume_timeout,
            reconnect_attempts,
            search_only,
            managed,
            disabled_sources,
            json.dumps(extras),
            json.dumps(yaml_data),
        )
        await self.invalidate_cache()

    async def get_connection_args(self) -> dict[str, int | str | bool | None, list[str]]:
        """Get the connection args for the node.

        Returns
        -------
        dict
            The connection args.
        """
        data = await self.fetch_all()

        if self.id in BUNDLED_NODES_IDS_HOST_MAPPING:
            data["yaml"]["lavalink"]["server"]["password"] = self.client._user_id

        return {
            "unique_identifier": self.id,
            "host": data["yaml"]["server"]["address"],
            "port": data["yaml"]["server"]["port"],
            "password": data["yaml"]["lavalink"]["server"]["password"],
            "name": data["name"],
            "ssl": data["ssl"],
            "reconnect_attempts": data["reconnect_attempts"],
            "search_only": data["search_only"],
            "resume_timeout": data["resume_timeout"],
            "disabled_sources": data["disabled_sources"],
            "managed": data["managed"],
        }

    @classmethod
    async def create_managed(cls, identifier: int) -> None:
        """Create the player in the database"""

        __, java_xmx_default, __, __ = get_jar_ram_actual(JAVA_EXECUTABLE)
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await NodeRow.raw(
            """
            INSERT INTO node
            (id, managed, ssl, reconnect_attempts, search_only, yaml, name, resume_key, resume_timeout, extras)
            VALUES ({}, {}, {}, {}, {}, {}, {}, {}, {}, {})
            ON CONFLICT (id) DO NOTHING;
            ;
            """,
            identifier,
            True,
            False,
            -1,
            False,
            json.dumps(NODE_DEFAULT_SETTINGS),
            "PyLavManagedNode",
            None,
            600,
            json.dumps({"max_ram": java_xmx_default}),
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from packaging.version import Version

# noinspection PyProtectedMember
from pylav._internals.pylav_yaml_builder import ENV_FILE
from pylav.constants.node import NODE_DEFAULT_SETTINGS
from pylav.constants.versions import VERSION_0_0_0, VERSION_1_0_0
from pylav.storage.migrations.logging import LOGGER

if TYPE_CHECKING:
    from pylav.core.client import Client


async def migration_v_1_0_0(client: Client, current_version: Version) -> None:
    if current_version >= VERSION_1_0_0 or current_version == VERSION_0_0_0:
        return

    LOGGER.info("Running %s migration", VERSION_1_0_0)
    config = client.node_db_manager.bundled_node_config()
    yaml_data = await config.fetch_yaml()
    yaml_data["lavalink"]["server"]["playerUpdateInterval"] = NODE_DEFAULT_SETTINGS["lavalink"]["server"][
        "playerUpdateInterval"
    ]
    yaml_data["lavalink"]["server"]["filters"] = NODE_DEFAULT_SETTINGS["lavalink"]["server"]["filters"]
    yaml_data["plugins"]["lavasrc"]["providers"] = list(
        dict.fromkeys(NODE_DEFAULT_SETTINGS["plugins"]["lavasrc"]["providers"])
    )
    yaml_data["plugins"]["lavasrc"]["applemusic"]["mediaAPIToken"] = NODE_DEFAULT_SETTINGS["plugins"]["lavasrc"][
        "applemusic"
    ]["mediaAPIToken"]
    yaml_data["logging"]["request"] = NODE_DEFAULT_SETTINGS["logging"]["request"]

    if yaml_data["plugins"]["lavasrc"]["spotify"]["clientId"] == "3d5cd36c73924786aa290798b2131c58":
        yaml_data["plugins"]["lavasrc"]["spotify"]["clientId"] = ""
        yaml_data["plugins"]["lavasrc"]["spotify"]["clientSecret"] = ""
        yaml_data["plugins"]["lavasrc"]["sources"]["spotify"] = False
        LOGGER.error(
            "You are using the default Spotify client ID and secret. "
            "Please register your own client ID and secret at "
            "https://developer.spotify.com/dashboard/applications "
            "and set them in your %s settings file.",
            ENV_FILE,
        )
    yaml_data["lavalink"]["server"]["resamplingQuality"] = NODE_DEFAULT_SETTINGS["lavalink"]["server"][
        "resamplingQuality"
    ]
    await config.update_yaml(yaml_data)
    await client.lib_db_manager.update_bot_dv_version(VERSION_1_0_0)



from __future__ import annotations

from pathlib import Path

import discord
from redbot.core.i18n import Translator
from redbot.core.utils.chat_formatting import box
from redbot.vendored.discord.ext import menus
from tabulate import tabulate

from pylav.extension.red.ui.menus.generic import BaseMenu
from pylav.helpers.format.ascii import EightBitANSI
from pylav.logging import getLogger
from pylav.type_hints.bot import DISCORD_COG_TYPE

LOGGER = getLogger("PyLav.ext.red.ui.sources.equalizer")
_ = Translator("PyLav", Path(__file__))


class EQPresetsSource(menus.ListPageSource):
    def __init__(self, cog: DISCORD_COG_TYPE, pages: list[tuple[str, dict]], per_page: int = 10):
        pages.sort()
        super().__init__(pages, per_page=per_page)
        self.cog = cog

    def get_starting_index_and_page_number(self, menu: BaseMenu) -> tuple[int, int]:
        page_num = menu.current_page
        start = page_num * self.per_page
        return start, page_num

    async def format_page(self, menu: BaseMenu, page: list[tuple[str, dict]]) -> discord.Embed:
        header_name = EightBitANSI.paint_yellow(_("Preset Name"), bold=True, underline=True)
        header_author = EightBitANSI.paint_yellow(_("Author"), bold=True, underline=True)
        data = []
        for preset_name, preset_data in page:
            try:
                author = self.cog.bot.get_user(preset_data["author"])
            except TypeError:
                author = _("Bundled with PyLav")
            data.append(
                {
                    header_name: EightBitANSI.paint_white(preset_name),
                    header_author: EightBitANSI.paint_blue(author),
                }
            )
        return await self.cog.pylav.construct_embed(
            messageable=menu.ctx, description=box(tabulate(data, headers="keys"), lang="ansi")
        )

    def get_max_pages(self):
        """:class:`int`: The maximum number of pages required to paginate this sequence"""
        return self._max_pages or 1



from __future__ import annotations

import asyncio
from pathlib import Path

from redbot.core.i18n import Translator

from pylav.core.context import PyLavContext
from pylav.extension.red.ui.menus.nodes import NodePickerMenu
from pylav.extension.red.ui.selectors.nodes import NodeSelectSelector
from pylav.extension.red.ui.sources.nodes import NodePickerSource
from pylav.helpers.format.strings import shorten_string
from pylav.nodes.node import Node
from pylav.type_hints.bot import DISCORD_COG_TYPE

_ = Translator("PyLav", Path(__file__))


async def maybe_prompt_for_node(cog: DISCORD_COG_TYPE, context: PyLavContext, nodes: list[Node]) -> Node | None:
    if len(nodes) > 1:
        node_picker = NodePickerMenu(
            cog=cog,
            bot=cog.bot,
            source=NodePickerSource(
                guild_id=context.guild.id,
                cog=cog,
                pages=nodes,
                message_str=shorten_string(
                    max_length=100, string=_("Multiple nodes matched. Pick the one which you meant.")
                ),
            ),
            selector_cls=NodeSelectSelector,
            delete_after_timeout=True,
            clear_buttons_after=True,
            starting_page=0,
            selector_text=shorten_string(max_length=100, string=_("Pick a node.")),
            original_author=context.interaction.user if context.interaction else context.author,
        )

        await node_picker.start(context)
        try:
            await node_picker.wait_for_response()
            node = node_picker.result
        except asyncio.TimeoutError:
            node = None
    else:
        node = nodes[0]
    return node



from __future__ import annotations

from redbot.core import commands

from pylav.core.context import PyLavContext


class IncompatibleException(EnvironmentError):
    """Command raised when an incompatible cog is already loaded"""


class MediaPlayerError(commands.CommandError):
    """
    Base class for all media player errors.
    """


class UnauthorizedChannelError(MediaPlayerError):
    """
    Raised when a command is used in a channel that is not authorized.
    """

    def __init__(self, channel: int):
        self.channel = channel


class MediaPlayerCheckError(MediaPlayerError):
    """
    Base class for all check errors.
    """


class MediaPlayerNotFoundError(MediaPlayerCheckError, commands.CheckFailure):
    """
    Raised when a media player is not found.
    """

    def __init__(self, context: PyLavContext) -> None:
        self.context = context


class NotDJError(MediaPlayerCheckError, commands.CheckFailure):
    """
    Raised when user does not have DJ Perms.
    """

    def __init__(self, context: PyLavContext) -> None:
        self.context = context



from __future__ import annotations

from pylav.players.filters.misc import FilterMixin


class Volume(FilterMixin):
    __slots__ = ("_value", "_default")

    def __init__(self, value: int | float | Volume) -> None:
        super().__init__()
        if isinstance(value, Volume):
            self.value = value.value
        elif isinstance(value, int):
            self.value = value / 100
        else:
            self.value = value

    def to_dict(self) -> dict[str, float | bool]:
        return {"volume": self.value}

    @classmethod
    def from_dict(cls, data: dict[str, float | bool]) -> Volume:
        return cls(data["volume"])

    def __float__(self) -> float:
        return self.value

    @property
    def value(self) -> float:
        return self._value

    @value.setter
    def value(self, v: float) -> None:
        if v < 0:
            raise ValueError(f"Volume must be must be greater than or equals to zero, not {v}")
        v = float(round(v, 2))
        self._value = v

    @classmethod
    def default(cls) -> Volume:
        return cls(value=1.0)

    def increase(self, by: float = 0.05) -> None:
        self.value += by

    def decrease(self, by: float = 0.05) -> None:
        self.value -= by

    def reset(self) -> None:
        self.value = 1.0

    def get(self) -> float:
        return self.value

    def get_int_value(self) -> int:
        return min(max(int(round(self.value * 100)), 0), 1000)

    def __repr__(self) -> str:
        return str(self.value)



from __future__ import annotations

import asyncio
import operator
import os
from functools import partial
from typing import TYPE_CHECKING

import aiohttp
import asyncstdlib

from pylav.compat import json
from pylav.constants.builtin_nodes import BUNDLED_NODES_IDS_HOST_MAPPING, PYLAV_BUNDLED_NODES_SETTINGS
from pylav.constants.config import EXTERNAL_UNMANAGED_NAME, JAVA_EXECUTABLE
from pylav.constants.coordinates import DEFAULT_REGIONS, REGION_TO_COUNTRY_COORDINATE_MAPPING
from pylav.events.node import NodeConnectedEvent, NodeDisconnectedEvent
from pylav.helpers.misc import ExponentialBackoffWithReset
from pylav.logging import getLogger
from pylav.nodes.node import Node
from pylav.nodes.utils import sort_key_nodes
from pylav.players.player import Player
from pylav.storage.models.node.mocked import NodeMock
from pylav.utils.location import get_closest_region_name_and_coordinate

if TYPE_CHECKING:
    from pylav.core.client import Client

LOGGER = getLogger("PyLav.NodeManager")


class NodeManager:
    __slots__ = (
        "_client",
        "_session",
        "_player_queue",
        "_unmanaged_external_host",
        "_unmanaged_external_password",
        "_unmanaged_external_port",
        "_unmanaged_external_ssl",
        "_nodes",
        "_adding_nodes",
        "_player_migrate_task",
    )

    def __init__(
        self,
        client: Client,
        external_host: str = None,
        external_password: str = None,
        external_port: int = None,
        external_ssl: bool = False,
    ):
        self._client = client
        self._session = aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=120), json_serialize=json.dumps)
        self._player_queue = set()
        self._unmanaged_external_host = external_host
        self._unmanaged_external_password = external_password
        self._unmanaged_external_port = external_port
        self._unmanaged_external_ssl = external_ssl
        self._nodes = []
        self._adding_nodes = asyncio.Event()
        self._player_migrate_task = None

    def __iter__(self):
        yield from self._nodes

    @property
    def session(self) -> aiohttp.ClientSession:
        return self._session

    @property
    def client(self) -> Client:
        """Returns the client"""
        return self._client

    @property
    def nodes(self) -> list[Node]:
        """Returns a list of all nodes"""
        return self._nodes

    @property
    def available_nodes(self) -> list[Node]:
        """Returns a list of available nodes"""
        return list(filter(operator.attrgetter("available"), self.nodes))

    @property
    def managed_nodes(self) -> list[Node]:
        """Returns a list of nodes that are managed by the client"""
        return list(filter(operator.attrgetter("managed"), self.nodes))

    @property
    def search_only_nodes(self) -> list[Node]:
        """Returns a list of nodes that are search only"""
        return list(filter(operator.attrgetter("available", "search_only"), self.nodes))

    @property
    def player_queue(self) -> list[Player]:
        """Returns a list of players that are queued to be played"""
        return list(self._player_queue)

    @player_queue.setter
    def player_queue(self, players: list[Player]) -> None:
        """Sets the player queue"""
        self._player_queue = set(players)

    @player_queue.deleter
    def player_queue(self):
        """Clears the player queue"""
        self._player_queue.clear()

    async def add_node(
        self,
        *,
        host: str,
        port: int,
        password: str,
        unique_identifier: int,
        name: str,
        resume_timeout: int = 60,
        reconnect_attempts: int = -1,
        ssl: bool = False,
        search_only: bool = False,
        disabled_sources: list[str] = None,
        managed: bool = False,
        yaml: dict | None = None,
        extras: dict = None,
        temporary: bool = False,
    ) -> Node:
        """
        Adds a node to PyLav's node manager.

        Parameters
        ----------
        host: :class:`str`
            The address of the Lavalink node.
        port: :class:`int`
            The port to use for websocket and REST connections.
        password: :class:`str`
            The password used for authentication.
        resume_timeout: Optional[:class:`int`]
            How long the node should wait for a connection while disconnected before clearing all players.
            Defaults to `60`.
        name: :class:`str`
            An identifier for the node that will show in logs. Defaults to `None`.
        reconnect_attempts: Optional[:class:`int`]
            The amount of times connection with the node will be reattempted before giving up.
            Set to `-1` for infinite. Defaults to `3`.
        ssl: Optional[:class:`bool`]
            Whether to use a ssl connection. Defaults to `False`.
        search_only: :class:`bool`
            Whether the node is search only. Defaults to `False`.
        unique_identifier: Optional[:class:`str`]
            A unique identifier for the node. Defaults to `None`.
        disabled_sources: Optional[:class:`list`[:class:`str`]]
            A list of sources to disable. Defaults to `None`.
        managed: Optional[:class:`bool`]
            Whether the node is managed by the client. Defaults to `False`.
        yaml: Optional[:class:`dict`]
            A dictionary of node settings. Defaults to `None`.
        extras: Optional[:class:`dict`]
            A dictionary of extra settings. Defaults to `{}`.
        temporary: :class:`bool`
            Whether the node is temporary. Defaults to `False`.
            Temporary nodes are not added to the db.

        Returns
        -------
        :class:`Node`
            The node that was added.
        """
        node = Node(
            manager=self,
            host=host,
            port=port,
            password=password,
            resume_timeout=resume_timeout,
            name=name,
            reconnect_attempts=reconnect_attempts,
            ssl=ssl,
            search_only=search_only,
            unique_identifier=unique_identifier,
            disabled_sources=disabled_sources,
            managed=managed,
            extras=extras or {},
            temporary=temporary,
        )
        self._nodes.append(node)

        # noinspection PyProtectedMember
        node._logger.info("Successfully added to Node Manager")
        # noinspection PyProtectedMember
        node._logger.verbose("Successfully added to Node Manager -- %r", node)

        if temporary:
            yaml = yaml or {"server": {}, "lavalink": {"server": {}}}
            yaml["server"]["address"] = host  # type: ignore
            yaml["server"]["port"] = port  # type: ignore
            yaml["lavalink"]["server"]["password"] = password
            data = {
                "name": name,
                "ssl": ssl,
                "resume_timeout": resume_timeout,
                "reconnect_attempts": reconnect_attempts,
                "search_only": search_only,
                "managed": managed,
                "extras": extras or {},
                "yaml": yaml,
                "disabled_sources": disabled_sources,
            }
            node._config = NodeMock(id=unique_identifier, data=data)

        else:
            node._config = await self.client.node_db_manager.update_node(
                host=host,
                port=port,
                password=password,
                resume_timeout=resume_timeout,
                name=name,
                reconnect_attempts=reconnect_attempts,
                ssl=ssl,
                search_only=search_only,
                unique_identifier=unique_identifier,
                disabled_sources=disabled_sources,
                managed=managed,
                yaml=yaml,
                extras=extras or {},
            )
        return node

    async def remove_node(self, node: Node) -> None:
        """
        Removes a node.
        Parameters
        ----------
        node: :class:`Node`
            The node to remove from the list.
        """
        await node.close()
        self.nodes.remove(node)
        # noinspection PyProtectedMember
        node._logger.info("Successfully removed Node")
        # noinspection PyProtectedMember
        node._logger.verbose("Successfully removed Node -- %r", node)
        if (
            node.identifier
            and not node.managed
            and node.identifier not in BUNDLED_NODES_IDS_HOST_MAPPING
            and node.identifier != 31415
        ):
            await self.client.node_db_manager.delete(node.identifier)
            # noinspection PyProtectedMember
            node._logger.debug("Successfully deleted Node from database")

    def get_region(self, endpoint: str | None) -> str | None:
        """
        Returns a region from a Discord voice server address.
        Parameters
        ----------
        endpoint: :class:`str`
            The address of the Discord voice server.
        Returns
        -------
        Optional[:class:`str`]
        """
        if not endpoint:
            return None

        endpoint = endpoint.replace("vip-", "")

        for key in DEFAULT_REGIONS:
            nodes = [n for n in self.available_nodes if n.region == key]

            if not nodes:
                continue

            if endpoint.startswith(key):
                return key
        return None

    def get_closest_node(self, region: str) -> Node:
        """
        Returns the closest node to a given region.
        Parameters
        ----------
        region: :class:`str`
            The region to use.
        Returns
        -------
        :class:`Node`
        """
        return min(self.available_nodes, key=lambda n: n.region_distance(region))

    async def find_best_node(
        self,
        region: str = None,
        not_region: str = None,
        feature: str = None,
        already_attempted_regions: set[str] = None,
        coordinates: tuple[float, float] = None,
        wait: bool = False,
        attempt: int = 0,
        backoff: ExponentialBackoffWithReset = None,
    ) -> Node | None:
        """Finds the best (least used) node in the given region, if applicable.
        Parameters
        ----------
        region: :class:`str`
            The region to use.
        not_region: :class:`str`
            The region to exclude.
        feature: :class:`str`
            The feature required.
        already_attempted_regions: :class:`set`[:class:`str`]
            A set of regions that have already been attempted.
        coordinates: :class:`tuple`[:class:`float`, :class:`float`]
            The coordinates to use.
        wait: :class:`bool`
            Whether to wait for a node to become available.
        attempt: :class:`int`
            The current attempt number.
        backoff: :class:`ExponentialBackoffWithReset`
            The backoff to use.
        Returns
        -------
        Optional[:class:`Node`]
        """
        if backoff is None:
            backoff = ExponentialBackoffWithReset()
            delay = 1
        else:
            delay = backoff.delay()
        already_attempted_regions = already_attempted_regions or set()
        if feature:
            nodes = [n for n in self.available_nodes if n.has_capability(feature)]
        else:
            nodes = self.available_nodes
        if coordinates is None:
            if region and region in REGION_TO_COUNTRY_COORDINATE_MAPPING:
                coordinates = REGION_TO_COUNTRY_COORDINATE_MAPPING[region]
            else:
                coordinates = (0, 0)
        if region and not_region:
            nodes = await self._get_nodes_by_region_with_exclusion(
                already_attempted_regions, coordinates, nodes, not_region, region
            )
        elif region:
            nodes = await self._get_nodes_by_region_only(already_attempted_regions, coordinates, nodes, region)
        else:
            nodes = [n for n in nodes if n.region != not_region and n.region not in already_attempted_regions]

        if not nodes:
            nodes = await self._get_fall_back_nodes(already_attempted_regions, feature, nodes)
        node = await asyncstdlib.min(nodes, key=partial(sort_key_nodes, region=region), default=None) if nodes else None
        if node is None and wait:
            await asyncio.sleep(delay)
            return await self.find_best_node(
                region=region,
                not_region=not_region,
                feature=feature,
                already_attempted_regions=already_attempted_regions,
                coordinates=coordinates,
                wait=wait,
                backoff=backoff,
                attempt=attempt + 1,
            )
        return node

    async def _get_fall_back_nodes(self, already_attempted_regions, feature, nodes):
        if feature:
            nodes = [
                n
                for n in self.available_nodes
                if n.has_capability(feature) and n.region not in already_attempted_regions
            ]
        else:
            nodes = self.available_nodes
        return nodes

    async def _get_nodes_by_region_only(self, already_attempted_regions, coordinates, nodes, region):
        available_regions = {n.region for n in self.available_nodes if n.region not in already_attempted_regions}
        closest_region, __ = await get_closest_region_name_and_coordinate(*coordinates, region_pool=available_regions)
        nodes = [
            n for n in nodes if (n.region in [region, closest_region]) and n.region not in already_attempted_regions
        ]
        return nodes

    async def _get_nodes_by_region_with_exclusion(
        self, already_attempted_regions, coordinates, nodes, not_region, region
    ):
        available_regions = {n.region for n in self.available_nodes if n.region not in already_attempted_regions}
        closest_region, __ = await get_closest_region_name_and_coordinate(*coordinates, region_pool=available_regions)
        nodes = [
            n
            for n in nodes
            if (n.region in [region, closest_region])
            and n.region != not_region
            and n.region not in already_attempted_regions
        ]
        return nodes

    def get_node_by_id(self, unique_identifier: int) -> Node | None:
        """
        Returns a node by its unique identifier.
        Parameters
        ----------
        unique_identifier: :class:`int`
            The unique identifier of the node.
        Returns
        -------
        Optional[:class:`Node`]
        """
        return next((n for n in self.nodes if n.identifier == unique_identifier), None)

    async def node_connect(self, node: Node) -> None:
        """
        Called when a node is connected from Lavalink.
        Parameters
        ----------
        node: :class:`Node`
            The node that has just connected.
        """
        # noinspection PyProtectedMember
        node._logger.debug("Successfully established connection")
        del node.down_votes
        self._player_migrate_task = asyncio.create_task(self._player_change_node_task(node))
        self.client.dispatch_event(NodeConnectedEvent(node))

    async def _player_change_node_task(self, node):
        for player in iter(self.player_queue):
            await player.change_node(node, forced=True)
            # noinspection PyProtectedMember
            node._logger.debug("Successfully moved %s", player.guild.id)
        # noinspection PyProtectedMember
        if self.client._connect_back:
            # noinspection PyProtectedMember
            for player in iter(node._original_players):
                await player.change_node(node, forced=True)
                player._original_node = None
        del self.player_queue
        self._player_migrate_task = None

    async def node_disconnect(self, node: Node, code: int, reason: str) -> None:
        """
        Called when a node is disconnected from Lavalink.
        Parameters
        ----------
        node: :class:`Node`
            The node that has just connected.
        code: :class:`int`
            The code for why the node was disconnected.
        reason: :class:`str`
            The reason why the node was disconnected.
        """
        if self.client.is_shutting_down:
            return
        # noinspection PyProtectedMember
        node._logger.warning("Disconnected with code %s and reason %s", code, reason)
        # noinspection PyProtectedMember
        node._logger.verbose(
            "Disconnected with code %s and reason %s -- %r",
            code,
            reason,
            node,
        )
        self.client.dispatch_event(NodeDisconnectedEvent(node, code, reason))
        best_node = await self.find_best_node(region=node.region)
        if not best_node or not best_node.available:
            self.player_queue = self.player_queue + node.players
            LOGGER.error("Unable to move players, no available nodes! Waiting for a node to become available")
            return

        for player in iter(node.players):
            await player.change_node(best_node, forced=True)
            # noinspection PyProtectedMember
            if self.client._connect_back:
                player._original_node = node

    async def close(self) -> None:
        if self._player_migrate_task is not None:
            self._player_migrate_task.cancel()
        await self.session.close()
        for node in iter(self.nodes):
            await node.close()

    async def connect_to_all_nodes(self) -> None:
        nodes_list = []
        for node in iter(await self.client.node_db_manager.get_all_unmanaged_nodes()):
            await self._process_single_unmanaged_node_connection(node, nodes_list)
        await self._process_envvar_node(nodes_list)
        # noinspection PyProtectedMember
        config_data = self.client._lib_config_manager.get_config()
        all_data = await config_data.fetch_all()

        if all_data["java_path"] != JAVA_EXECUTABLE and os.path.exists(JAVA_EXECUTABLE):
            await config_data.update_java_path(JAVA_EXECUTABLE)

        if all_data["use_bundled_pylav_external"]:
            # await self._process_bundled_node_london(nodes_list)
            await self._process_bundled_node_ny(nodes_list)

        tasks = [asyncio.create_task(n.wait_until_ready()) for n in nodes_list]
        if not tasks:
            if await self.client.managed_node_is_enabled():
                self._adding_nodes.set()
                return
            LOGGER.warning("No nodes found, please add some nodes")
            return
        done, pending = await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
        for task in pending:
            task.cancel()
        for result in done:
            result.result()
        len_nodes = sum(1 for node in nodes_list if node.available)
        if len_nodes == 0:
            raise OSError("No nodes are available")
        if not self._adding_nodes.is_set():
            self._adding_nodes.set()

    async def _process_bundled_node_lava_link(self, nodes_list):
        if all(True for n in iter(nodes_list) if n.host != "lava.link"):
            nodes_list.append(
                await self.add_node(
                    password=f"PyLav/{self.client.lib_version}",
                    **PYLAV_BUNDLED_NODES_SETTINGS["lava.link"],
                )
            )
        else:
            LOGGER.debug(
                "%s already added to connection pool - skipping duplicated connection",
                PYLAV_BUNDLED_NODES_SETTINGS["lava.link"]["name"],
            )

    async def _process_bundled_node_ny(self, nodes_list):
        if all(True for n in iter(nodes_list) if n.host != "ll-us-ny.draper.wtf") and not self.get_node_by_id(
            PYLAV_BUNDLED_NODES_SETTINGS["ll-us-ny.draper.wtf"]["unique_identifier"]
        ):
            base_settings = PYLAV_BUNDLED_NODES_SETTINGS["ll-us-ny.draper.wtf"]
            nodes_list.append(await self.add_node(**base_settings))
        else:
            LOGGER.debug(
                "%s already added to connection pool - skipping duplicated connection",
                PYLAV_BUNDLED_NODES_SETTINGS["ll-us-ny.draper.wtf"]["name"],
            )

    async def _process_bundled_node_london(self, nodes_list):
        if all(True for n in iter(nodes_list) if n.host != "ll-gb.draper.wtf") and not self.get_node_by_id(
            PYLAV_BUNDLED_NODES_SETTINGS["ll-gb.draper.wtf"]["unique_identifier"]
        ):
            base_settings = PYLAV_BUNDLED_NODES_SETTINGS["ll-gb.draper.wtf"]
            base_settings["host"] = "ll-gb.draper.wtf"
            nodes_list.append(await self.add_node(**base_settings))
        else:
            LOGGER.debug(
                "%s already added to connection pool - skipping duplicated connection",
                PYLAV_BUNDLED_NODES_SETTINGS["ll-gb.draper.wtf"]["name"],
            )

    async def _process_envvar_node(self, nodes_list):
        if self._unmanaged_external_host and self._unmanaged_external_password:
            if all(True for n in nodes_list if n.host != self._unmanaged_external_host):
                if self._unmanaged_external_host in PYLAV_BUNDLED_NODES_SETTINGS:
                    base_settings = PYLAV_BUNDLED_NODES_SETTINGS[self._unmanaged_external_host]
                else:
                    base_settings = {
                        "port": self._unmanaged_external_port or (443 if self._unmanaged_external_ssl else 80),
                        "ssl": self._unmanaged_external_ssl,
                        "password": self._unmanaged_external_password,
                        "resume_timeout": 600,
                        "reconnect_attempts": -1,
                        "search_only": False,
                        "managed": False,
                        "disabled_sources": [],
                        "host": self._unmanaged_external_host,
                        "unique_identifier": 31415,
                        "name": EXTERNAL_UNMANAGED_NAME,
                        "temporary": True,
                    }
                nodes_list.append(await self.add_node(**base_settings))
            else:
                LOGGER.warning(
                    "%s already added to connection pool - skipping duplicated connection - (%s:%s)",
                    EXTERNAL_UNMANAGED_NAME,
                    self._unmanaged_external_host,
                    self._unmanaged_external_port,
                )

    async def _process_single_unmanaged_node_connection(self, node, nodes_list):
        if node.id == self.client.bot.user.id:
            LOGGER.debug("Skipping node %s as it is the managed node", node.id)
            return
        node_data = await node.fetch_all()
        try:
            if node in nodes_list:
                LOGGER.warning(
                    "%s Node already added to connection pool - skipping duplicated connection - (%s:%s)",
                    node_data["name"],
                    node_data["yaml"]["server"]["address"],
                    node_data["yaml"]["server"]["port"],
                )
                return
            if node_data["yaml"]["server"]["address"] in PYLAV_BUNDLED_NODES_SETTINGS:
                connection_arguments = PYLAV_BUNDLED_NODES_SETTINGS[node_data["yaml"]["server"]["address"]]
            else:
                connection_arguments = await node.get_connection_args()
            nodes_list.append(await self.add_node(**connection_arguments))
        except (ValueError, KeyError) as exc:
            LOGGER.warning("Invalid node, skipping ... id: %s - Original error: %s", node.id, exc)

    async def wait_until_ready(self, timeout: float | None = None):
        await asyncio.wait_for(self._adding_nodes.wait(), timeout=timeout)



from __future__ import annotations

from collections.abc import Callable
from typing import Any

from pylav.type_hints.bot import DISCORD_COG_TYPE, DISCORD_COMMAND_ERROR_TYPE, DISCORD_CONTEXT_TYPE
from pylav.type_hints.generics import CORO_TYPE, MaybeCoro

DISCORD_CHECK_TYPE = Callable[[DISCORD_CONTEXT_TYPE], MaybeCoro[bool]]

DISCORD_HOOK_TYPE = (
    Callable[[DISCORD_COG_TYPE, DISCORD_CONTEXT_TYPE], CORO_TYPE[Any]]
    | Callable[[DISCORD_CONTEXT_TYPE], CORO_TYPE[Any]]
)

DISCORD_ERROR_TYPE = (
    Callable[[DISCORD_COG_TYPE, DISCORD_CONTEXT_TYPE, DISCORD_COMMAND_ERROR_TYPE], CORO_TYPE[Any]]
    | Callable[[DISCORD_CONTEXT_TYPE, DISCORD_COMMAND_ERROR_TYPE], CORO_TYPE[Any]],
)



from __future__ import annotations

from piccolo.columns import JSONB, UUID, Array, BigInt, Boolean, Integer
from piccolo.table import Table

from pylav.storage.database.tables.misc import DATABASE_ENGINE


class PlayerRow(Table, db=DATABASE_ENGINE, tablename="player"):
    primary_key = UUID(primary_key=True)
    id = BigInt(index=True)
    bot = BigInt(index=True, null=False)
    volume = Integer(null=False, default=100)
    max_volume = Integer(null=False, default=1000)
    auto_play_playlist_id = BigInt(null=False, default=1)

    text_channel_id = BigInt(null=True, default=0)
    notify_channel_id = BigInt(null=True, default=0)
    forced_channel_id = BigInt(null=True, default=0)

    repeat_current = Boolean(null=False, default=False)
    repeat_queue = Boolean(null=False, default=False)
    shuffle = Boolean(null=False, default=True)
    auto_shuffle = Boolean(null=False, default=True)
    auto_play = Boolean(null=False, default=True)
    self_deaf = Boolean(null=False, default=True)
    empty_queue_dc = JSONB(
        null=False,
        default={
            "enabled": False,
            "time": 60,
        },
    )
    alone_dc = JSONB(
        null=False,
        default={
            "enabled": False,
            "time": 60,
        },
    )
    alone_pause = JSONB(
        null=False,
        default={
            "enabled": False,
            "time": 60,
        },
    )
    extras = JSONB(null=False, default={})
    effects = JSONB(null=False, default={})
    dj_users = Array(null=False, default=[], base_column=BigInt())
    dj_roles = Array(null=False, default=[], base_column=BigInt())



from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, Literal, TypeAlias, Union  # noqa

from pylav.exceptions.request import HTTPException
from pylav.nodes.api.responses.exceptions import LoadException
from pylav.nodes.api.responses.filters import Filters
from pylav.nodes.api.responses.misc import Git, Plugin, Version
from pylav.nodes.api.responses.playlists import Info
from pylav.nodes.api.responses.shared import PlaylistPluginInfo
from pylav.nodes.api.responses.track import Track
from pylav.nodes.api.responses.websocket import CPU, Frame, Memory
from pylav.type_hints.dict_typing import JSON_DICT_TYPE


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class BaseTrackResponse:
    playlistInfo: Info | None
    pluginInfo: PlaylistPluginInfo | None
    exception: LoadException | None
    tracks: list[Track]

    def __post_init__(self):
        if self.pluginInfo is None:
            object.__setattr__(self, "pluginInfo", PlaylistPluginInfo(kwargs=None))


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class TrackLoaded(BaseTrackResponse):  # noqa
    loadType: Literal["TRACK_LOADED"]


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class PlaylistLoaded(BaseTrackResponse):  # noqa
    loadType: Literal["PLAYLIST_LOADED"]


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class SearchResult(BaseTrackResponse):  # noqa
    loadType: Literal["SEARCH_RESULT"]


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class NoMatches(BaseTrackResponse):  # noqa
    loadType: Literal["NO_MATCHES"]


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class LoadFailed(BaseTrackResponse):  # noqa
    loadType: Literal["LOAD_FAILED"]


LoadTrackResponses: TypeAlias = Union[TrackLoaded, PlaylistLoaded, NoMatches, LoadFailed, SearchResult, HTTPException]


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class LavalinkInfo:
    version: Version
    buildTime: int
    git: Git
    jvm: str
    lavaplayer: str
    sourceManagers: list[str]
    filters: list[str]
    plugins: list[Plugin]

    def __post_init__(self):
        temp = []
        for s in self.plugins:
            if isinstance(s, Plugin) or (isinstance(s, dict) and (s := Plugin(**s))):
                temp.append(s)
        object.__setattr__(self, "plugins", temp)


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class VoiceState:
    token: str
    endpoint: str
    sessionId: str
    connected: bool | None = None
    ping: int | None = -1

    def to_dict(self) -> JSON_DICT_TYPE:
        return {
            "token": self.token,
            "endpoint": self.endpoint,
            "sessionId": self.sessionId,
            "connected": self.connected,
            "ping": self.ping,
        }

    def __repr__(self) -> str:
        return (
            f"<VoiceStateObject(token={'OBFUSCATED' if self.token else None} "
            f"endpoint={self.endpoint} "
            f"sessionId={self.sessionId} "
            f"connected={self.connected} "
            f"ping={self.ping})"
        )


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class LavalinkPlayer:
    guildId: str
    volume: int
    paused: bool
    voice: VoiceState | dict
    filters: Filters | dict
    track: Track | dict | None = None

    def __post_init__(self) -> None:
        if isinstance(self.voice, dict):
            object.__setattr__(self, "voice", VoiceState(**self.voice))
        if isinstance(self.filters, dict):
            object.__setattr__(self, "filters", Filters(**self.filters))
        if isinstance(self.track, dict):
            object.__setattr__(self, "track", Track(**self.track))

    def to_dict(
        self,
    ) -> JSON_DICT_TYPE:
        assert (
            isinstance(self.voice, VoiceState)
            and isinstance(self.filters, Filters)
            and isinstance(self.track, (Track, type(None)))
        )
        return {
            "guildId": self.guildId,
            "volume": self.volume,
            "paused": self.paused,
            "voice": self.voice.to_dict(),
            "filters": self.filters.to_dict(),
            "track": self.track.to_dict() if self.track else None,
        }


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Stats:
    players: int
    playingPlayers: int
    uptime: int
    memory: Memory
    cpu: CPU
    frameStats: Frame | None = None
    uptime_seconds: int = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        object.__setattr__(self, "uptime_seconds", self.uptime / 1000)



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

from packaging.version import Version

from pylav.constants.playlists import BUNDLED_PYLAV_PLAYLISTS_IDS
from pylav.constants.versions import VERSION_0_0_0, VERSION_0_8_5_0
from pylav.storage.migrations.logging import LOGGER

if TYPE_CHECKING:
    from pylav.core.client import Client


async def migration_v_0_8_5_0(client: Client, current_version: Version) -> None:
    if current_version >= VERSION_0_8_5_0 or current_version == VERSION_0_0_0:
        return

    LOGGER.info("Running %s migration", VERSION_0_8_5_0)
    playlists = [
        p for p in await client.playlist_db_manager.get_bundled_playlists() if p.id in BUNDLED_PYLAV_PLAYLISTS_IDS
    ]
    for playlist in playlists:
        await playlist.delete()
    t = asyncio.create_task(client.playlist_db_manager.update_bundled_playlists(*BUNDLED_PYLAV_PLAYLISTS_IDS))
    t.set_name(f"Update bundled playlists for migration {VERSION_0_8_5_0}")
    # noinspection PyProtectedMember
    client._update_schema_manager._tasks_depend_on_node.append(t)
    await client.lib_db_manager.update_bot_dv_version(VERSION_0_8_5_0)



from __future__ import annotations

from pathlib import Path

import discord
from redbot.core.i18n import Translator

from pylav.helpers.format.strings import shorten_string
from pylav.logging import getLogger
from pylav.type_hints.bot import DISCORD_COG_TYPE, DISCORD_INTERACTION_TYPE

LOGGER = getLogger("PyLav.ext.red.ui.modals.queue")
_ = Translator("PyLav", Path(__file__))


class EnqueueModal(discord.ui.Modal):
    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        title: str,
        timeout: float | None = None,
    ):
        super().__init__(title=title, timeout=timeout)
        self.cog = cog
        self.text = discord.ui.TextInput(
            style=discord.TextStyle.paragraph,
            label=shorten_string(max_length=100, string=_("Search for a song to add to the queue.")),
            placeholder=shorten_string(
                max_length=100,
                string="Hello by Adele, speak:Hello, https://open.spotify.com/playlist/37i9dQZF1DX6XceWZP1znY",
            ),
        )
        self.add_item(self.text)

    async def on_submit(self, interaction: DISCORD_INTERACTION_TYPE):
        await self.cog.command_play.callback(
            self.cog,
            interaction,
            query=self.text.value.strip(),
        )



from __future__ import annotations

import contextlib
from pathlib import Path
from typing import TYPE_CHECKING

import discord
import humanize
from redbot.core import i18n
from redbot.core.i18n import Translator
from redbot.core.utils.chat_formatting import box, humanize_number
from redbot.vendored.discord.ext import menus
from tabulate import tabulate

from pylav.extension.red.ui.selectors.options.nodes import NodeOption
from pylav.helpers.format.ascii import EightBitANSI
from pylav.logging import getLogger
from pylav.nodes.node import Node
from pylav.type_hints.bot import DISCORD_COG_TYPE

if TYPE_CHECKING:
    from pylav.extension.red.ui.menus.nodes import NodeManagerMenu, NodePickerMenu


LOGGER = getLogger("PyLav.ext.red.ui.sources.node")
_ = Translator("PyLav", Path(__file__))


class NodePickerSource(menus.ListPageSource):
    def __init__(self, guild_id: int, cog: DISCORD_COG_TYPE, pages: list[Node], message_str: str):
        super().__init__(entries=pages, per_page=5)
        self.message_str = message_str
        self.per_page = 5
        self.guild_id = guild_id
        self.select_options: list[NodeOption] = []
        self.cog = cog
        self.select_mapping: dict[str, Node] = {}

    def get_starting_index_and_page_number(self, menu: NodePickerMenu) -> tuple[int, int]:
        page_num = menu.current_page
        start = page_num * self.per_page
        return start, page_num

    async def format_page(self, menu: NodePickerMenu, nodes: list[Node]) -> discord.Embed | str:
        idx_start, page_num = self.get_starting_index_and_page_number(menu)
        page = await self.cog.pylav.construct_embed(messageable=menu.ctx, title=self.message_str)

        total_number_of_entries = len(self.entries)
        current_page = humanize_number(page_num + 1)
        total_number_of_pages = humanize_number(self.get_max_pages())

        match total_number_of_entries:
            case 1:
                message = _("Page 1 / 1 | 1 entry")
            case 0:
                message = _("Page 1 / 1 | 0 entries")
            case __:
                message = _(
                    "Page {current_page_variable_do_not_translate} / {total_number_of_pages_variable_do_not_translate} | {total_number_of_entries_variable_do_not_translate} entries"
                ).format(
                    current_page_variable_do_not_translate=current_page,
                    total_number_of_pages_variable_do_not_translate=total_number_of_pages,
                    total_number_of_entries_variable_do_not_translate=humanize_number(total_number_of_entries),
                )
        page.set_footer(text=message)
        return page

    async def get_page(self, page_number):
        if page_number > self.get_max_pages():
            page_number = 0
        base = page_number * self.per_page
        self.select_options.clear()
        self.select_mapping.clear()
        for i, node in enumerate(self.entries[base : base + self.per_page], start=base):  # noqa: E203
            self.select_options.append(await NodeOption.from_node(node=node, index=i))
            self.select_mapping[f"{node.id}"] = node
        return self.entries[base : base + self.per_page]  # noqa: E203

    def get_max_pages(self):
        """:class:`int`: The maximum number of pages required to paginate this sequence"""
        return self._max_pages or 1


class NodeListSource(menus.ListPageSource):
    def __init__(self, cog: DISCORD_COG_TYPE, pages: list[Node]):
        super().__init__(entries=pages, per_page=1)
        self.cog = cog

    def get_starting_index_and_page_number(self, menu: NodeManagerMenu) -> tuple[int, int]:
        page_num = menu.current_page
        start = page_num * self.per_page
        return start, page_num

    async def format_page(self, menu: NodeManagerMenu, node: Node) -> discord.Embed | str:
        locale = f"{i18n.get_babel_locale()}"
        with contextlib.suppress(Exception):
            humanize.i18n.activate(locale)
        idx_start, page_num = self.get_starting_index_and_page_number(menu)
        region = node.region
        coord = node.coordinates
        host = node.host
        port = node.port
        password = node.password
        no = EightBitANSI.paint_red(_("No"))
        yes = EightBitANSI.paint_green(_("Yes"))

        secure = yes if node.ssl else no
        connected = yes if node.available else no
        search_only = yes if node.search_only else no

        node_stats = await node.fetch_stats()
        pylav_connected_players = len(node.connected_players)
        pylav_active_players = len(node.playing_players)

        if node_stats:
            server_connected_players = node_stats.players
            server_active_players = node_stats.playingPlayers
            if node.stats:
                frames_sent = node.stats.frames_sent
                frames_nulled = node.stats.frames_nulled
                frames_deficit = node.stats.frames_deficit
            else:
                frames_sent = 0
                frames_nulled = 0
                frames_deficit = 0
            uptime = humanize.naturaldelta(node_stats.uptime_seconds)
            system_load = humanize_number(round(node_stats.cpu.systemLoad, 2))
            lavalink_load = humanize_number(round(node_stats.cpu.lavalinkLoad, 2))
            free = humanize.naturalsize(node_stats.memory.free, binary=True)
            used = humanize.naturalsize(node_stats.memory.used, binary=True)
            allocated = humanize.naturalsize(node_stats.memory.allocated, binary=True)
            reservable = humanize.naturalsize(node_stats.memory.reservable, binary=True)
            penalty = humanize_number(round(node.penalty - 1, 2))
        else:
            server_connected_players = 0
            server_active_players = 0
            frames_sent = 0
            frames_nulled = 0
            frames_deficit = 0
            uptime = "?"
            system_load = "?"
            lavalink_load = "?"
            free = "?"
            used = "?"
            allocated = "?"
            reservable = "?"
            penalty = "?"
        feature_str = ""
        for feature in sorted(node.capabilities):
            feature_str += f"{EightBitANSI.paint_blue(feature)}\n"
        feature_str = feature_str.strip() or EightBitANSI.paint_red(_("None / Unknown"))
        humanize.i18n.deactivate()
        t_property = EightBitANSI.paint_yellow(_("Property"), bold=True, underline=True)
        t_values = EightBitANSI.paint_yellow(_("Value"), bold=True, underline=True)
        coordinate_str = EightBitANSI.paint_white(
            _("Latitude: {latitude_variable_do_not_translate}\nLongitude: {longitude_variable_do_not_translate}")
        ).format(
            latitude_variable_do_not_translate=EightBitANSI.paint_blue(coord[0] if coord else "?"),
            longitude_variable_do_not_translate=EightBitANSI.paint_blue(coord[1] if coord else "?"),
        )

        data = {
            EightBitANSI.paint_white(_("Region")): EightBitANSI.paint_blue(region or _("N/A")),
            EightBitANSI.paint_white(_("Coordinates")): coordinate_str,
            EightBitANSI.paint_white(_("Host")): EightBitANSI.paint_blue(host),
            EightBitANSI.paint_white(_("Port")): EightBitANSI.paint_blue(port),
            EightBitANSI.paint_white(_("Password")): EightBitANSI.paint_blue("*" * min([len(password), 10])),
            EightBitANSI.paint_white(_("SSL")): secure,
            EightBitANSI.paint_white(_("Available")): connected,
            EightBitANSI.paint_white(_("Search Only")): search_only,
            EightBitANSI.paint_white(_("Players\nConnected\nActive")): EightBitANSI.paint_blue(
                f"-\n{pylav_connected_players}/{server_connected_players or '?'}\n"
                f"{pylav_active_players}/{server_active_players or '?'}"
            ),
            EightBitANSI.paint_white(_("Frames Lost")): EightBitANSI.paint_blue(
                f"{(abs(frames_deficit) + abs(frames_nulled))/(frames_sent or 1) * 100:.2f}%"
            )
            if ((abs(frames_deficit) + abs(frames_nulled)) / (frames_sent or 1)) > 0
            else EightBitANSI.paint_blue("0%"),
            EightBitANSI.paint_white(_("Uptime")): EightBitANSI.paint_blue(uptime),
            EightBitANSI.paint_white(_("CPU Load\nLavalink\nSystem")): EightBitANSI.paint_blue(
                f"-\n{lavalink_load}%\n{system_load}%"
            ),
            EightBitANSI.paint_white(_("Penalty")): EightBitANSI.paint_blue(penalty),
            EightBitANSI.paint_white(_("Memory\nUsed\nFree\nAllocated\nReservable")): EightBitANSI.paint_blue(
                f"-\n{used}\n{free}\n{allocated}\n{reservable}"
            ),
            EightBitANSI.paint_white(_("Features")): feature_str,
        }
        description = box(
            tabulate([{t_property: k, t_values: v} for k, v in data.items()], headers="keys", tablefmt="fancy_grid"),
            lang="ansi",
        )
        embed = await self.cog.pylav.construct_embed(
            messageable=menu.ctx,
            title=node.name,
            description=description,
        )

        total_number_of_entries = len(self.entries)
        current_page = humanize_number(page_num + 1)
        total_number_of_pages = humanize_number(self.get_max_pages())
        match total_number_of_entries:
            case 1:
                message = _("Page 1 / 1 | 1 node")
            case 0:
                message = _("Page 1 / 1 | 0 nodes")
            case __:
                message = _(
                    "Page {current_page_variable_do_not_translate} / {total_number_of_pages_variable_do_not_translate} | {total_number_of_entries_variable_do_not_translate} nodes"
                ).format(
                    current_page_variable_do_not_translate=current_page,
                    total_number_of_pages_variable_do_not_translate=total_number_of_pages,
                    total_number_of_entries_variable_do_not_translate=humanize_number(total_number_of_entries),
                )

        embed.set_footer(text=message)
        return embed


class NodeManageSource(menus.ListPageSource):
    target: Node | None

    def __init__(self, cog: DISCORD_COG_TYPE):
        self.cog = cog
        self.target = None
        self.per_page = 1

    @property
    def entries(self) -> list[Node]:
        return self.cog.pylav.node_manager.nodes

    @property
    def _max_pages(self) -> int:
        return len(self.entries) or 1

    async def get_page(self, page_number: int) -> Node:
        """Returns either a single element of the sequence or
        a slice of the sequence.

        If :attr:`per_page` is set to ``1`` then this returns a single
        element. Otherwise it returns at most :attr:`per_page` elements.

        Returns
        ---------
        Node
            The data returned.
        """
        self.target = self.entries[page_number]
        return self.target

    def get_starting_index_and_page_number(self, menu: NodeManagerMenu) -> tuple[int, int]:
        page_num = menu.current_page
        start = page_num * self.per_page
        return start, page_num

    async def format_page(self, menu: NodeManagerMenu, node: Node) -> discord.Embed | str:
        locale = f"{i18n.get_babel_locale()}"
        with contextlib.suppress(Exception):
            humanize.i18n.activate(locale)
        await menu.prepare()
        idx_start, page_num = self.get_starting_index_and_page_number(menu)
        region = node.region
        coord = node.coordinates
        host = node.host
        port = node.port
        password = node.password
        no = EightBitANSI.paint_red(_("No"))
        yes = EightBitANSI.paint_green(_("Yes"))

        secure = yes if node.ssl else no
        connected = yes if node.available else no
        search_only = yes if node.search_only else no

        node_stats = await node.fetch_stats()
        pylav_connected_players = len(node.connected_players)
        pylav_active_players = len(node.playing_players)

        if node_stats:
            server_connected_players = node_stats.players
            server_active_players = node_stats.playingPlayers
            if node.stats:
                frames_sent = node.stats.frames_sent
                frames_nulled = node.stats.frames_nulled
                frames_deficit = node.stats.frames_deficit
            else:
                frames_sent = 0
                frames_nulled = 0
                frames_deficit = 0
            uptime = humanize.naturaldelta(node_stats.uptime_seconds)
            system_load = humanize_number(round(node_stats.cpu.systemLoad, 2))
            lavalink_load = humanize_number(round(node_stats.cpu.lavalinkLoad, 2))
            free = humanize.naturalsize(node_stats.memory.free, binary=True)
            used = humanize.naturalsize(node_stats.memory.used, binary=True)
            allocated = humanize.naturalsize(node_stats.memory.allocated, binary=True)
            reservable = humanize.naturalsize(node_stats.memory.reservable, binary=True)
            penalty = humanize_number(round(node.penalty - 1, 2))
        else:
            server_connected_players = 0
            server_active_players = 0
            frames_sent = 0
            frames_nulled = 0
            frames_deficit = 0
            uptime = "?"
            system_load = "?"
            lavalink_load = "?"
            free = "?"
            used = "?"
            allocated = "?"
            reservable = "?"
            penalty = "?"
        feature_str = ""
        for feature in sorted(node.capabilities):
            feature_str += f"{EightBitANSI.paint_blue(feature)}\n"
        feature_str = feature_str.strip() or EightBitANSI.paint_red(_("None / Unknown"))
        humanize.i18n.deactivate()
        t_property = EightBitANSI.paint_yellow(_("Property"), bold=True, underline=True)
        t_values = EightBitANSI.paint_yellow(_("Value"), bold=True, underline=True)
        coordinate_str = EightBitANSI.paint_white(
            _("Latitude: {latitude_variable_do_not_translate}\nLongitude: {longitude_variable_do_not_translate}")
        ).format(
            latitude_variable_do_not_translate=EightBitANSI.paint_blue(coord[0] if coord else "?"),
            longitude_variable_do_not_translate=EightBitANSI.paint_blue(coord[1] if coord else "?"),
        )

        data = {
            EightBitANSI.paint_white(_("Region")): EightBitANSI.paint_blue(region or _("N/A")),
            EightBitANSI.paint_white(_("Coordinates")): coordinate_str,
            EightBitANSI.paint_white(_("Host")): EightBitANSI.paint_blue(host),
            EightBitANSI.paint_white(_("Port")): EightBitANSI.paint_blue(port),
            EightBitANSI.paint_white(_("Password")): EightBitANSI.paint_blue("*" * min([len(password), 10])),
            EightBitANSI.paint_white(_("SSL")): secure,
            EightBitANSI.paint_white(_("Available")): connected,
            EightBitANSI.paint_white(_("Search Only")): search_only,
            EightBitANSI.paint_white(_("Players\nConnected\nActive")): EightBitANSI.paint_blue(
                f"-\n{pylav_connected_players}/{server_connected_players or '?'}\n"
                f"{pylav_active_players}/{server_active_players or '?'}"
            ),
            EightBitANSI.paint_white(_("Frames Lost")): EightBitANSI.paint_blue(
                f"{(abs(frames_deficit) + abs(frames_nulled)) / (frames_sent or 1) * 100:.2f}%"
            )
            if ((abs(frames_deficit) + abs(frames_nulled)) / (frames_sent or 1)) > 0
            else EightBitANSI.paint_blue("0%"),
            EightBitANSI.paint_white(_("Uptime")): EightBitANSI.paint_blue(uptime),
            EightBitANSI.paint_white(_("CPU Load\nLavalink\nSystem")): EightBitANSI.paint_blue(
                f"-\n{lavalink_load}%\n{system_load}%"
            ),
            EightBitANSI.paint_white(_("Penalty")): EightBitANSI.paint_blue(penalty),
            EightBitANSI.paint_white(_("Memory\nUsed\nFree\nAllocated\nReservable")): EightBitANSI.paint_blue(
                f"-\n{used}\n{free}\n{allocated}\n{reservable}"
            ),
            EightBitANSI.paint_white(_("Features")): feature_str,
        }
        description = box(
            tabulate([{t_property: k, t_values: v} for k, v in data.items()], headers="keys", tablefmt="fancy_grid"),
            lang="ansi",
        )
        embed = await self.cog.pylav.construct_embed(
            messageable=menu.ctx,
            title=node.name,
            description=description,
        )

        number_of_pages = self.get_max_pages()
        total_number_of_entries = len(self.entries)
        current_page = humanize_number(page_num + 1)
        total_number_of_pages = humanize_number(self.get_max_pages())
        match number_of_pages:
            case 1:
                message = _("Page 1 / 1 | 1 node")
            case 0:
                message = _("Page 1 / 1 | 0 nodes")
            case __:
                message = _(
                    "Page {current_page_variable_do_not_translate} / {total_number_of_pages_variable_do_not_translate} | {total_number_of_entries_variable_do_not_translate} nodes"
                ).format(
                    current_page_variable_do_not_translate=current_page,
                    total_number_of_pages_variable_do_not_translate=total_number_of_pages,
                    total_number_of_entries_variable_do_not_translate=total_number_of_entries,
                )

        embed.set_footer(text=message)
        return embed



from __future__ import annotations

from discord.app_commands import AppCommandError
from discord.ext.commands import CommandError


class PyLavException(CommandError, AppCommandError):
    """Base exception for errors in the library"""
